/*

*/

import renderer.scene.*;
import renderer.models.Disk;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.ComponentEvent;

/**
  Run this program and then resize the window so that is no longer
  a square window. Notice how the circle is distorted into an ellipse.
<p>
  When you resize the window, the aspect ratio of the view rectangle in
  view  coordinates (that is, the aspect ratio of the near plane of the
  camera's view volume) no longer matches the aspect ration of the
  framebuffer's viewport. The view rectangle gets distorted so that it
  fits into the viewport.
<p>
  We can fix this by having Java's resize window event handler create
  a new viewport within the FrameBuffer every time there is a "reshape"
  event and  set the viewport to always have the same aspect ratio as
  the camera's view rectangle (see the next program, Circle_v2.java).
*/
public class Circle_v1_Distort extends Circle_v0_Abstract
{
   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public Circle_v1_Distort()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      // Set up the camera's view volume.
      final double left   = -1.0;
      final double right  =  1.0;
      final double bottom = -1.0;
      final double top    =  1.0;
      scene.camera.projOrtho(left, right, bottom, top);

      // Create a model of a circle.
      final Model model = new Disk(1.0, 10, 40);

      ModelShading.setRandomColor(model);

      // Add the Model to the Scene.
      scene.addModel(model);


      // Define initial dimensions for a FrameBuffer.
      final int width  = SIZE;
      final int height = SIZE;

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.darkGray);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      final JFrame jf = new JFrame("Renderer 6: Distort");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
    //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      if (showFBaspectRatio)
         System.out.printf("Aspect ratio (of framebuffer) = %.2f\n", (double)w/(double)h);

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport().getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(w, h, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);

      setupViewing();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new Circle_v1_Distort()
      );
   }
}
