/*

*/

import renderer.scene.*;
import renderer.models.Disk;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.ComponentEvent;

/**
   This example demonstrates a form of "letterboxing".
<p>
   This program uses Java's resize window event handler and
   the FrameBuffer's viewport to solve the problem of changing
   aspect ratios demonstrated by the program Circle_v1.java
<p>
   This program always uses a square viewport (which matches the
   shape of the camera's view rectangle), which is as large as
   possible and is always positioned in the lower left hand corner
   of this program's FrameBuffer.
<p>
   As an exercise, modify this program to always place the
   square viewport in the upper left hand corner of the
   FrameBuffer.
<p>
   Also try placing the square viewport centered on the left
   edge of the FrameBuffer.
<p>
   Finally, try placing the square viewport at the center of
   the FrameBuffer (true letterboxing).
*/
public class Circle_v2_Letterbox extends Circle_v0_Abstract
{
   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public Circle_v2_Letterbox()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      // Set up the camera's view volume.
      final double left   = -1.0;
      final double right  =  1.0;
      final double bottom = -1.0;
      final double top    =  1.0;
      scene.camera.projOrtho(left, right, bottom, top);

      // Create a model of a circle.
      final Model model = new Disk(1.0, 10, 40);

      ModelShading.setRandomColor(model);

      // Add the Model to the Scene.
      scene.addModel(model);


      // Define initial dimensions for a FrameBuffer.
      final int width  = SIZE;
      final int height = SIZE;

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.darkGray);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      final JFrame jf = new JFrame("Renderer 6: Letterbox");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
    //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      if (showFBaspectRatio)
         System.out.printf("Aspect ratio (of framebuffer) = %.2f\n", (double)w/(double)h);

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport().getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(w, h, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);

      setupViewing();
   }


   // Get in one place the code to set up the viewport and the view volume.
   protected void setupViewing()
   {
      // Get the size of the FrameBuffer.
      final FrameBuffer fb = fbp.getFrameBuffer();
      final int w = fb.width;
      final int h = fb.height;

      final int d = Math.min(w, h); // dimension of the viewport
      final int hOffset = (h < w) ? (w - h) / 2 : 0;
      final int vOffset = (w < h) ? (h - w) / 2 : 0;
      // Fit the largest possible viewport, with aspect ratio 1,
      // in the framebuffer, located at the framebuffer's:
      if (mode == 1)
      {
         fb.setViewport(0,       0,       d, d); // 1. upper left hand corner
      }
      else if (mode == 2)
      {
         fb.setViewport(hOffset, 0,       d, d); // 2. top edge, centered
      }
      else if (mode == 3)
      {
         fb.setViewport(w - d,   0,       d, d); // 3. upper right hand corner
      }
      else if (mode == 4)
      {
         fb.setViewport(w - d,   vOffset, d, d); // 4. right edge, centered
      }
      else if (mode == 5)
      {
         fb.setViewport(w - d,   h - d,   d, d); // 5, lower right hand corner
      }
      else if (mode == 6)
      {
         fb.setViewport(hOffset, h - d,   d, d); // 6. bottom edge, centered
      }
      else if (mode == 7)
      {
         fb.setViewport(0,       h - d,   d, d); // 7. lower left hand corner
      }
      else if (mode == 8)
      {
         fb.setViewport(0,       vOffset, d, d); // 8. left edge, centered
      }
      else if (mode == 9)
      {
         fb.setViewport(hOffset, vOffset, d, d); // 9. centered in the framebuffer
      }

    //fb.setViewport(0,       0,       d, d); // 1. upper left hand corner
    //fb.setViewport(w - d,   0,       d, d); // 2. upper right hand corner
    //fb.setViewport(0,       h - d,   d, d); // 3. lower left hand corner
    //fb.setViewport(w - d,   h - d,   d, d); // 4, lower right hand corner

    //fb.setViewport(0,       vOffset, d, d); // 5. left, center
    //fb.setViewport(hOffset, 0,       d, d); // 6. top, center
    //fb.setViewport(w - d,   vOffset, d, d); // 7. right, center
    //fb.setViewport(hOffset, h - d,   d, d); // 8. bottom, center

    //fb.setViewport(hOffset, vOffset, d, d); // 9. centered in the framebuffer

      if (showCamera)
      {
         System.out.print( scene.camera );
      }

      // Render again.
      fb.clearFB();
      fb.vp.clearVP();
      Pipeline.render(scene, fb);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new Circle_v2_Letterbox()
      );
   }
}
