/*

*/

import renderer.scene.*;
import renderer.models.Disk;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.ComponentEvent;

/**
  The example demonstrates a form of "cropping".
<p>
  This program keeps the aspect ratio and the size of the complete
  circle at a constant 512 by 512 pixels.

  If the program window is smaller than 512 by 512 pixels, then only part of the complete circle is
  shown in the program window
  (our view of the scene gets "cropped" to fit inside of the screen window).

  If the program window is smaller than 512 by 512 pixels, then only part of the complete circle is
  shown in the program window
  (the view of our circle in the camera's view rectangle gets "cropped" to fit inside of the screen window).

  Also, the complete circle is anchored in the upper left hand corner of this program's window.
  But the program window can be any size.

<p>
  Run this program and resize the window so that it is smaller than
  512 by 512 and see what happens.
<p>
  As the user changes the shape (the aspect ratio) and size of the
  screen window, the code in the componentResized() event handler
  needs to change the size and location of the framebuffer's viewport
  and also the size and location of the camera's view volume.
<p>
  We create this effect in the componentResized() event handler by
  using the FrameBuffer's setViewport() method and the Camera's
  projOrtho() method.
<p>
  As an exercise, modify this program to always have the complete circle
  anchored in the lower right hand corner of this program's window. Try
  resizing the window using either the window's lower right hand corner
  or the window's upper left hand corner.
<p>
  As an exercise, modify this program to always have the complete circle
  centered at the center of the screen window (that is, centered in the
  FrameBuffer).
*/
public class Circle_v3_Crop extends Circle_v0_Abstract
{
   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public Circle_v3_Crop()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      // Set up the camera's view volume.
      final double left   = -1.0;
      final double right  =  1.0;
      final double bottom = -1.0;
      final double top    =  1.0;
      scene.camera.projOrtho(left, right, bottom, top);

      // Create a model of a circle.
      final Model model = new Disk(1.0, 10, 40);

      ModelShading.setRandomColor(model);

      // Add the Model to the Scene.
      scene.addModel(model);


      // Define initial dimensions for a FrameBuffer.
      final int width  = SIZE;
      final int height = SIZE;

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.darkGray);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      final JFrame jf = new JFrame("Renderer 6: Crop");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }



   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
    //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      final int wFB = fbp.getWidth();
      final int hFB = fbp.getHeight();

      if (showFBaspectRatio)
         System.out.printf("Aspect ratio (of framebuffer) = %.2f\n", (double)wFB/(double)hFB);

      if (0 == hFB) return; // there is no window to draw

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport().getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(wFB, hFB, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);

      setupViewing();
   }


   // Get in one place the code to set up the viewport and the view volume.
   protected void setupViewing()
   {
      // Get the size of the FrameBuffer.
      final FrameBuffer fb = fbp.getFrameBuffer();
      final int wFB = fb.width;
      final int hFB = fb.height;

      // Set the size of the viewport.
      final int wVP = (wFB < SIZE) ? wFB : SIZE;
      final int hVP = (hFB < SIZE) ? hFB : SIZE;

      if (mode == 1)
      {
         // 1. Upper left hand corner.
         fb.setViewport(0, 0, wVP, hVP);
         scene.camera.projOrtho(-1,                       // left
                                -1 + (2.0 * wVP) / SIZE,  // right
                                 1 - (2.0 * hVP) / SIZE,  // bottom
                                 1);                      // top
      }
      else if (mode == 2)
      {
         // 2. Centered on top edge.
         fb.setViewport((wFB - wVP)/2, 0, wVP, hVP);
         scene.camera.projOrtho(-wVP / (double)SIZE,      // left
                                 wVP / (double)SIZE,      // right
                                 1 - (2.0 * hVP) / SIZE,  // bottom
                                 1);                      // top
      }
      else if (mode == 3)
      {
         // 3. Upper right hand corner.
         fb.setViewport(wFB - wVP, 0, wVP, hVP);
         scene.camera.projOrtho( 1 - (2.0 * wVP) / SIZE,  // left
                                 1,                       // right
                                 1 - (2.0 * hVP) / SIZE,  // bottom
                                 1);                      // top
      }
      else if (mode == 4)
      {
         // 4. Centered on right edge.
         fb.setViewport(wFB - wVP, (hFB - hVP)/2, wVP, hVP);
         scene.camera.projOrtho( 1 - (2.0 * wVP) / SIZE,  // left
                                 1,                       // right
                                -hVP / (double)SIZE,      // bottom
                                 hVP / (double)SIZE);     // top
      }
      else if (mode == 5)
      {
         // 5. Lower right hand corner.
         fb.setViewport(wFB - wVP, hFB - hVP, wVP, hVP);
         scene.camera.projOrtho( 1 - (2.0 * wVP) / SIZE,  // left
                                 1,                       // right
                                -1,                       // bottom
                                -1 + (2.0 * hVP) / SIZE); // top
      }
      else if (mode == 6)
      {
         // 6. Centered on bottom edge.
         fb.setViewport((wFB - wVP)/2, hFB - hVP, wVP, hVP);
         scene.camera.projOrtho(-wVP / (double)SIZE,      // left
                                 wVP / (double)SIZE,      // right
                                -1,                       // bottom
                                -1 + (2.0 * hVP) / SIZE); // top
      }
      else if (mode == 7)
      {
         // 7. Lower left hand corner.
         fb.setViewport(0, hFB - hVP, wVP, hVP);
         scene.camera.projOrtho(-1,                       // left
                                -1 + (2.0 * wVP) / SIZE,  // right
                                -1,                       // bottom
                                -1 + (2.0 * hVP) / SIZE); // top
      }
      else if (mode == 8)
      {
         // 8. Centered on left edge.
         fb.setViewport(0, (hFB - hVP)/2, wVP, hVP);
         scene.camera.projOrtho(-1,                       // left
                                -1 + (2.0 * wVP) / SIZE,  // right
                                -hVP / (double)SIZE,      // bottom
                                 hVP / (double)SIZE);     // top
      }
      else if (mode == 9)
      {
         // 9. Centered
         fb.setViewport((wFB - wVP)/2, (hFB - hVP)/2, wVP, hVP);
         scene.camera.projOrtho(-wVP / (double)SIZE,   // left
                                 wVP / (double)SIZE,   // right
                                -hVP / (double)SIZE,   // bottom
                                 hVP / (double)SIZE);  // top
      }

      if (showCamera)
      {
         System.out.print( scene.camera );
      }

      // Render again.
      fb.clearFB();
      fb.vp.clearVP();
      Pipeline.render(scene, fb);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new Circle_v3_Crop()
      );
   }
}
