/*

*/

import renderer.scene.*;
import renderer.models.Disk;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.ComponentEvent;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.Scrollbar;

/**
  The example demonstrates a form of "pan-and-scan".
<p>
  This program keeps the aspect ratio and the size of the complete
  circle at a constant 512 by 512 pixels. The program window can be
  any size. If the program window is smaller than 512 by 512 pixels,
  then only part of the complete circle is shown in the program window
  (the view of our circle in the camera's view rectangle gets "cropped"
  to fit inside of the screen window).
*/
public class Circle_v5_Pan_and_Scan extends Circle_v0_Abstract
                                    implements AdjustmentListener
{
   final private Scrollbar sbV;
   final private Scrollbar sbH;
   private int sliderValueH = 50;
   private int sliderValueV = 50;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public Circle_v5_Pan_and_Scan()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      // Set up the camera's view volume.
      final double left   = -1.0;
      final double right  =  1.0;
      final double bottom = -1.0;
      final double top    =  1.0;
      scene.camera.projOrtho(left, right, bottom, top);

      // Create a model of a circle.
      final Model model = new Disk(1.0, 10, 40);

      ModelShading.setRandomColor(model);

      // Add the Model to the Scene.
      scene.addModel(model);


      // Define initial dimensions for a FrameBuffer.
      final int width  = SIZE;
      final int height = SIZE;

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.darkGray);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      final JFrame jf = new JFrame("Renderer 6: Pan and Scan");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      // Create Scrollbars for the JFrame.
      sbV = new Scrollbar(Scrollbar.VERTICAL,   50, 2, 0, 100);
      sbH = new Scrollbar(Scrollbar.HORIZONTAL, 50, 2, 0, 100);
      // Place the Scrollbars in the JFrame.
      jf.getContentPane().add(sbV, BorderLayout.EAST );
      jf.getContentPane().add(sbH, BorderLayout.SOUTH );
      jf.pack();
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
      sbV.addAdjustmentListener(this);
      sbH.addAdjustmentListener(this);
   }



   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
    //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      final int wFB = fbp.getWidth();
      final int hFB = fbp.getHeight();

      if (showFBaspectRatio)
         System.out.printf("Aspect ratio (of framebuffer) = %.2f\n", (double)wFB/(double)hFB);

      if (0 == hFB) return; // there is no window to draw

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport().getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(wFB, hFB, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);

      draw();
   }


   // Implement the AdjustmentListener interface for the scrollbar.
   @Override public void adjustmentValueChanged(AdjustmentEvent e)
   {
    //super.adjustmentValueChanged(e);
      sliderValueV = sbV.getValue();
      sliderValueH = sbH.getValue();
      System.out.println("slider_H = " + sliderValueH + ", "
                          + "slider_V = " + sliderValueV);
      draw();
   }


   // Render the current Scene into the FrameBuffer's viewport.
   void draw()
   {
      // Get the size of the FrameBuffer.
      final FrameBuffer fb = this.fbp.getFrameBuffer();
      final int w = fb.width;
      final int h = fb.height;

      // Create a viewport.
      int vpX = 0;
      int vpY = 0;
      int vpW = w;
      int vpH = h;
      if (w > SIZE)
      {
         vpW = SIZE;
         vpX = (w - SIZE)/2;
      }
      if (h > SIZE)
      {
         vpH = SIZE;
         vpY = (h - SIZE)/2;
      }
      fb.setViewport(vpX, vpY, vpW, vpH);
      fb.clearFB();
      fb.vp.clearVP();

      // Create a view volume whose aspect ratio matches the
      // viewport's aspect ratio.
      double left   = -(vpW/(double)SIZE);
      double right  =  (vpW/(double)SIZE);
      double bottom = -(vpH/(double)SIZE);
      double top    =  (vpH/(double)SIZE);
      // Use the slider value to pan the view volume across the scene.
      left   += (sliderValueH - 50.0)/50.0 * (1.0 - vpW/(double)SIZE);
      right  += (sliderValueH - 50.0)/50.0 * (1.0 - vpW/(double)SIZE);
      bottom += (50.0 - sliderValueV)/50.0 * (1.0 - vpH/(double)SIZE);
      top    += (50.0 - sliderValueV)/50.0 * (1.0 - vpH/(double)SIZE);
      scene.camera.projOrtho(left, right, bottom, top);

      Pipeline.render(scene, fb);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new Circle_v5_Pan_and_Scan()
      );
   }
}
