/*

*/

package renderer.pipeline;
import  renderer.scene.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Optional;
import java.awt.Color;

/**
   Clip a (projected) {@link LineSegment} that sticks out of the
   view rectangle in the image plane.
<p>
   This clipping algorithm is a simplification of the Liang-Barsky
   Parametric Line Clipping algorithm.
<p>
   This algorithm assumes that all {@code Vertex} objects have been
   projected onto the {@link renderer.scene.Camera}'s image plane,
   {@code z = -1}. This algorithm also assumes that the camera's view
   rectangle in the image plane is
   <pre>{@code
      -1 <= x <= +1  and
      -1 <= y <= +1.
   }</pre>
<p>
   If a line segment's projected vertex has an {@code x} or {@code y}
   coordinate with absolute value greater than 1, then that vertex
   "sticks out" of the view rectangle. This algorithm will clip the
   line segment so that both of the line segment's vertices are within
   the view rectangle.
<p>
   Here is an outline of the clipping algorithm.
<p>
   Recursively process each line segment, using the following steps.
<p>
     1) Test if the line segment no longer needs to be clipped, i.e.,
        both of its vertices are within the view rectangle. If this is
        the case, then return the line segment wrapped in an
        {@link Optional} object.
        <pre>{@code
               x=-1        x=+1
                 |          |
                 |          |
             ----+----------+----- y = +1
                 |     v1   |
                 |    /     |
                 |   /      |
                 |  /       |
                 | v0       |
             ----+----------+----- y = -1
                 |          |
                 |          |
        }</pre>
<p>
     2) Test if the line segment should be "trivially rejected". A line
        segment is "trivially rejected" if it is on the wrong side of any
        of the four lines that bound the view rectangle (i.e., the four
        lines {@code x = 1}, {@code x = -1}, {@code y = 1}, {@code y = -1}).
        If so, then return an empty {@link Optional} object indicating that
        the line segment is not be rasterized into the framebuffer.
<p>
        Notice that a line like the following one is trivially rejected
        because it is on the "wrong" side of the line {@code x = 1}.
        <pre>{@code
                           x=1
                            |            v1
                            |            /
                 +----------+           /
                 |          |          /
                 |          |         /
                 |          |        /
                 |          |       /
                 |          |      /
                 +----------+     /
                            |    /
                            |  v0
        }</pre>
        But the following line is NOT trivially rejected because, even
        though it is completely outside of the view rectangle, this line
        is not entirely on the wrong side of any one of the four lines
        {@code x = 1}, {@code x = -1}, {@code y = 1}, or {@code y = -1}.
        The line below will get clipped at least one time (either on the
        line {@code x = 1} or the line {@code y = -1}) before it is
        (recursively) a candidate for "trivial rejection". Notice that
        the line below could even be clipped twice, first on {@code y = 1},
        then on {@code x = 1}, before it can be trivially rejected (by
        being on the wrong side of {@code y = -1}).
        <pre>{@code
                           x=1
                            |          v1
                            |         /
                 +----------+        /
                 |          |       /
                 |          |      /
                 |          |     /
                 |          |    /
                 |          |   /
                 +----------+  /
                            | /
                            |/
                            /
                           /|
                          / |
                        v0
        }</pre>
<p>
     3) If the line segment has been neither accepted nor rejected, then
        it needs to be clipped. So we test the line segment against each
        of the four clipping lines, {@code x = 1}, {@code x = -1},
        {@code y = 1}, and {@code y = -1}, to determine if the line segment
        crosses one of those lines. We clip the line segment against the
        first line which we find that it crosses. Then we recursively clip
        the resulting clipped line segment. Notice that we only clip against
        the first clipping line which the segment is found to cross. We do
        not continue to test against the other clipping lines. This is
        because it may be the case, after just one clip, that the line
        segment is now a candidate for trivial accept or reject. So rather
        than test the line segment against several more clipping lines
        (which may be useless tests) it is more efficient to recursively
        clip the line segment, which will then start with the trivial accept
        or reject tests.
<p>
        When we clip a line segment against a clipping line, it is always
        the case that one endpoint of the line segment is on the "right"
        side of the clipping line and the other endpoint is on the "wrong"
        side of the clipping line. In the following picture, assume that
        {@code v0} is on the "wrong" side of the clipping line({@code x = 1})
        and {@code v1} is on the "right" side. So {@code v0} needs to be
        clipped off the line segment and replaced by a new vertex.
        <pre>{@code
                             x=1
                         v1   |
                           \  |
                            \ |
                             \|
                              \
                              |\
                              | \
                              |  \
                              |   v0
        }</pre>
        Represent points {@code p(t)} on the line segment between {@code v0}
        and {@code v1} with the following parametric equation.
        <pre>{@code
                  p(t) = (1-t) * v0 + t * v1  with  0 <= t <= 1
        }</pre>
        Notice that this equation parameterizes the line segment starting
        with {@code v0} at {@code t=0} (on the "wrong side") and ending
        with {@code v1} at {@code t=1} (on the "right side"). We need to
        find the value of {@code t} when the line segment crosses the
        clipping line {@code x = 1}. Let {@code v0 = (x0, y0)} and let
        {@code v1 = (x1, y1)}. Then the above parametric equation becomes
        the two component equations
        <pre>{@code
                 x(t) = (1-t) * x0 + t * x1,
                 y(t) = (1-t) * y0 + t * y1,  with  0 <= t <= 1.
        }</pre>
        Since the clipping line in this example is {@code x = 1}, we need
        to solve the equation {@code x(t) = 1} for {@code t}. So we need
        to solve
        <pre>{@code
                  1 = (1-t) * x0 + t * x1
        }</pre>
        for {@code t}. Here are a few algebra steps.
        <pre>{@code
                  1 = x0 - t * x0 + t * x1
                  1 = x0 + t * (x1 - x0)
                  1 - x0 = t * (x1 - x0)
                       t = (1 - x0)/(x1 - x0)
        }</pre>
        We get similar equations for {@code t} if we clip against the other
        clipping lines ({@code x = -1}, {@code y = 1}, or {@code y = -1})
        and we assume that {@code v0} is on the "wrong side" and {@code v1}
        is on the "right side".
<p>
        Let {@code t0} denote the above value for {@code t}. With this value
        for {@code t}, we can compute the y-coordinate of the new vertex
        {@code p(t0)} that replaces {@code v0}.
        <pre>{@code
                             x=1
                        v1    |
                          \   |
                           \  |
                            \ |
                              p(t0)=(1, y(t0))
                              |
                              |
                              |
         }</pre>
         Here is the algebra.
         <pre>{@code
                  y(t0) = (1-t0) * y0 + t0 * y1
                        = y0 + t0 * (y1 - y0)
                        = y0 + (1 - x0)*((y1 - y0)/(x1 - x0))
         }</pre>
         Finally, the new line segment between {@code v1} and the new
         vertex {@code p(t0)} is recursively clipped so that it can be
         checked to see if it should be trivially accepted, trivially
         rejected, or clipped again.
*/
public final class Clip
{
   public static boolean debug = false;

   /**
      Start with a {@link Model} that contains {@link LineSegment}s
      that have been projected onto the camera's view plane, {@code z = -1}.
   <p>
      If a projected {@link LineSegment} sticks out of the camera's
      view rectangle, then replace that {@link LineSegment}, in the
      {@link Model}'s list of line segments, with one that has been
      clipped so that it is contained in the view rectangle.
   <p>
      If a projected {@link LineSegment} is completely outside of
      the view rectangle, then drop that {@link LineSegment} from
      the {@link Model}'s list of line segments.
   <p>
      Return a {@link Model} for which every {@link LineSegment} is
      completely contained in the camera's view rectangle.

      @param model  {@link Model} containing projected {@link LineSegment}s
      @return a {@link Model} containing {@link LineSegment}s clipped to the view rectangle
   */
   public static Model clip(final Model model)
   {
      boolean debug = Clip.debug && (Pipeline.debug || model.debug);

      final List<LineSegment> newLineSegmentList = new ArrayList<>();

      // Replace the model's original list of vertices and colors
      // with shallow copies.
      final Model model2 =  new Model(new ArrayList<>(model.vertexList),
                                      model.lineSegmentList,
                                      new ArrayList<>(model.colorList),
                                      model.name,
                                      model.visible,
                                      model.debug);

      for (final LineSegment ls : model2.lineSegmentList)
      {
         Pipeline.logLineSegment("2. Clipping", model2, ls);

         Optional<LineSegment> ls_clipped = clip(model2, ls);

         if ( ls_clipped.isPresent() )
         {
            // Keep the line segments that are visible.
            newLineSegmentList.add( ls_clipped.get() );

            Pipeline.logLineSegment("2. Clipping (accept)", model2, ls_clipped.get());
         }
         else
         {
            // Discard the line segments that are not visible.
            Pipeline.logLineSegment("2. Clipping (reject)", model2, ls);
         }
      }
      // Replace the model's original list of line segments
      // with the list of clipped line segments.
      return new Model(model2.vertexList, // has been updated with clipped vertices
                       newLineSegmentList,
                       model2.colorList,  // has been updated with interpolated colors
                       model2.name,
                       model2.visible,
                       model2.debug);
   }


   /**
      If the {@link LineSegment} sticks out of the view rectangle,
      then return a clipped version that is contained in the view
      rectangle. The new, clipped {@link LineSegment} object is
      returned wrapped in an {@link Optional} object.
      <p>
      At least one new clipped {@link Vertex} will be added to the
      {@link Model}'s vertex list (and as many as four new vertices
      may be added to the {@link Model}'s vertex list).
      <p>
      If the {@link LineSegment} is completely outside of the view
      rectangle, then return an empty {@link Optional} object to
      indicate that the {@link LineSegment} should be discarded.

      @param model  {@link Model} that the {@link LineSegment} {@code ls} comes from
      @param ls     {@link LineSegment} to be clipped
      @return a clipped version of {@code ls} wrapped in an {@link Optional} object
   */
   private static Optional<LineSegment> clip(final Model model, final LineSegment ls)
   {
      boolean debug = Clip.debug && (Pipeline.debug || model.debug);

      // Make local copies of several values.
      final Vertex v0 = model.vertexList.get( ls.vIndex[0] );
      final Vertex v1 = model.vertexList.get( ls.vIndex[1] );

      final double x0 = v0.x,  y0 = v0.y;
      final double x1 = v1.x,  y1 = v1.y;

      // 1. Check for trivial accept.
      if ( ! ( Math.abs( x0 ) > 1
            || Math.abs( y0 ) > 1
            || Math.abs( x1 ) > 1
            || Math.abs( y1 ) > 1 ) )
      {
         if (debug) System.err.println("-Trivial accept.");
         return Optional.of(ls); // better than "return ls;"
      }
      // 2. Check for trivial delete.
      else if ( (x0 >  1 && x1 >  1)   // to the right of the line x = 1
             || (x0 < -1 && x1 < -1)   // to the left of the line x = -1
             || (y0 >  1 && y1 >  1)   // above the line y = 1
             || (y0 < -1 && y1 < -1) ) // below the line y = -1
      {
         if (debug) System.err.println("-Trivial delete.");
         return Optional.empty(); // better than "return null;"
      }
      // 3. Need to clip this line segment.
      else
      {
         // recursively clip a new (clipped) line segment
         return clip(model, clipOneTime(model, ls));
      }
   }


   /**
      This method takes in a line segment that cosses one of the four
      clipping lines. This method returns a new LineSegment that uses
      a new, interpolated, Vertex (with a new interpolated Color) that
      is the intersection point between the given LineSegment and the
      crossed clipping line.
      <p>
      This method solves for the value of {@code t} for which the
      parametric equation
      <pre>{@code
                  p(t) = (1-t) * v_outside + t * v_inside
      }</pre>
      intersects the crossed clipping line. (Notice that the equation
      is parameterized so that we move from the outside vertex towards
      the inside vertex as {@code t} increases from 0 to 1.) The solved
      for value of {@code t} is then plugged into the parametric formula
      to get the coordinates of the intersection point and also to compute
      the new interpolated color components.
      <p>
      The new interpolated Vertex and Color objects are added to the end
      of the vertex and color lists from the given Model.

      @param model  {@link Model} that the {@link LineSegment} {@code ls} comes from
      @param ls     the {@link LineSegment} being clipped
      @return a new clipped {@link LineSegment} object
   */
   private static LineSegment clipOneTime(
                                 final Model model,
                                 final LineSegment ls)
   {
      boolean debug = Clip.debug && (Pipeline.debug || model.debug);

      // Make local copies of several values.
      final Vertex v0 = model.vertexList.get( ls.vIndex[0] );
      final Vertex v1 = model.vertexList.get( ls.vIndex[1] );

      final double x0 = v0.x,  y0 = v0.y;
      final double x1 = v1.x,  y1 = v1.y;

      final double vIx, vIy;  // "I" for "inside"
      final double vOx, vOy;  // "O" for "outside"
      final int eqnNum;  // keep track of which edge is crossed
      final double t;    // when we cross the clipping line
      final double x;    // x-coordinate of where we cross the clipping line
      final double y;    // y-coordinate of where we cross the clipping line
      final String equation;
      final String vOutside;
      final int vIndex;  // index for newly interpolated Vertex
      final int inside;  // keep track of which vertex is on the inside

      if (x0 > 1)  // ls crosses the line x = 1
      {
         inside = 1;
         equation = "x = +1";
         vOutside = "v0";
         vOx = x0;  vOy = y0;
         vIx = x1;  vIy = y1;
         t = (1 - vOx) / (vIx - vOx);
         x = 1;
         y = (1-t) * vOy + t * vIy;
         Vertex newVertex = new Vertex(x, y, 0);
         // Modify the Model to contain the new Vertex.
         vIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
      }
      else if (x1 > 1)  // ls crosses the line x = 1
      {
         inside = 0;
         equation = "x = +1";
         vOutside = "v1";
         vIx = x0;  vIy = y0;
         vOx = x1;  vOy = y1;
         t = (1 - vOx) / (vIx - vOx);
         x = 1;  // prevent rounding errors
         y = (1-t) * vOy + t * vIy;
         final Vertex newVertex = new Vertex(x, y, 0);
         vIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
      }
      else if (x0 < -1)  // ls crosses the line x = -1
      {
         inside = 1;
         equation = "x = -1";
         vOutside = "v0";
         vOx = x0;  vOy = y0;
         vIx = x1;  vIy = y1;
         t = (-1 - vOx) / (vIx - vOx);
         x = -1;  // prevent rounding errors
         y = (1-t) * vOy + t * vIy;
         final Vertex newVertex = new Vertex(x, y, 0);
         vIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
      }
      else if (x1 < -1)  // ls crosses the line x = -1
      {
         inside = 0;
         equation = "x = -1";
         vOutside = "v1";
         vIx = x0;  vIy = y0;
         vOx = x1;  vOy = y1;
         t = (-1 - vOx) / (vIx - vOx);
         x = -1;  // prevent rounding errors
         y = (1-t) * vOy + t * vIy;
         final Vertex newVertex = new Vertex(x, y, 0);
         vIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
      }
      else if (y0 > 1)  // ls crosses the line y = 1
      {
         inside = 1;
         equation = "y = +1";
         vOutside = "v0";
         vOx = x0;  vOy = y0;
         vIx = x1;  vIy = y1;
         t = (1 - vOy) / (vIy - vOy);
         x = (1-t) * vOx + t * vIx;
         y = 1;  // prevent rounding errors
         final Vertex newVertex = new Vertex(x, y, 0);
         vIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
      }
      else if (y1 > 1)  // ls crosses the line y = 1
      {
         inside = 0;
         equation = "y = +1";
         vOutside = "v1";
         vIx = x0;  vIy = y0;
         vOx = x1;  vOy = y1;
         t = (1 - vOy) / (vIy - vOy);
         x = (1-t) * vOx + t * vIx;
         y = 1;  // prevent rounding errors
         final Vertex newVertex = new Vertex(x, y, 0);
         vIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
      }
      else if (y0 < -1)  // ls crosses the line y = -1
      {
         inside = 1;
         equation = "y = -1";
         vOutside = "v0";
         vOx = x0;  vOy = y0;
         vIx = x1;  vIy = y1;
         t = (-1 - vOy) / (vIy - vOy);
         x = (1-t) * vOx + t * vIx;
         y = -1;  // prevent rounding errors
         final Vertex newVertex = new Vertex(x, y, 0);
         vIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
      }
      else // if (y1 < -1)  // ls crosses the line y = -1
      {
         inside = 0;
         equation = "y = -1";
         vOutside = "v1";
         vIx = x0;  vIy = y0;
         vOx = x1;  vOy = y1;
         t = (-1 - vOy) / (vIy - vOy);
         x = (1-t) * vOx + t * vIx;
         y = -1;  // prevent rounding errors
         final Vertex newVertex = new Vertex(x, y, 0);
         vIndex = model.vertexList.size();
         model.vertexList.add(newVertex);
      }

      // Use the value of t to interpolate the color of the new vertex.
      float cI[] = model.colorList.get( ls.cIndex[inside]   ).getRGBColorComponents(null);
      float cO[] = model.colorList.get( ls.cIndex[1-inside] ).getRGBColorComponents(null);
      final float t_ = (float)t;
      final float r = (1-t_) * cO[0] + t_ * cI[0];
      final float g = (1-t_) * cO[1] + t_ * cI[1];
      final float b = (1-t_) * cO[2] + t_ * cI[2];

      // Modify the Model to contain the new Color.
      final Color newColor = new Color(r, g, b);
      final int cIndex = model.colorList.size();
      model.colorList.add(newColor);

      if (debug)
      {
         System.err.printf("-Clip off %s at %s\n", vOutside, equation);
         System.err.printf("- t = % .25f\n", t);
         System.err.printf("- <x_i,y_i> = <% .24f % .24f\n",  vIx, vIy);
         System.err.printf("- <x_o,y_o> = <% .24f % .24f\n",  vOx, vOy);
         System.err.printf("- <x_c,y_c> = <% .24f % .24f\n",    x,   y);
         System.err.printf("- <r_i,g_i,b_i> = <% .15f  % .15f  % .15f>\n",
                              cI[0], cI[1], cI[2]);
         System.err.printf("- <r_o,g_o,b_o> = <% .15f  % .15f  % .15f>\n",
                              cO[0], cO[1], cO[2]);
         System.err.printf("- <r_c,g_c,b_c> = <% .15f  % .15f  % .15f>\n",
                               r,  g,  b);
      }

      // Return a new LineSegment using the new Vertex and Color
      // and keeping the old LineSegment's inside Vertex and Color.
      final LineSegment newLS;
      if (0 == inside)
      {
         newLS = new LineSegment(ls.vIndex[0], vIndex, ls.cIndex[0], cIndex);
      }
      else
      {
         newLS = new LineSegment(vIndex, ls.vIndex[1], cIndex, ls.cIndex[1]);
      }
      return newLS;
   }
}
