/*

*/

package renderer.scene;

import java.util.List;
import java.util.ArrayList;
import java.awt.Color;

/**
   A {@code Model} data structure represents a distinct geometric object
   in a {@link Scene}. A {@code Model} data structure is mainly a {@link List}
   of {@link Vertex} objects, a list of {@link LineSegment} objects, and a
   list of {@link Color} objects. Each {@link LineSegment} object contains
   four integers that are the indices of two {@link Vertex} objects from the
   {@code Model}'s vertex list and two {@link Color} objects from the
   {@code Model}'s color list. The two {@link Vertex} objects contain the
   coordinates, in the camera coordinate system, for each of the line segment's
   two endpoints. The two {@link Color} objects contain the rgb values for
   each of the line segment's two endpoints.
<p>
   A {@code Model} represent the geometric object as a "wire-frame" of line
   segments, that is, the geometric object is drawn as a collection of "edges".
   This is a fairly simplistic way of doing 3D graphics and we will
   improve this in later renderers.
<p>
   See
<br> <a href="http://en.wikipedia.org/wiki/Wire-frame_model" target="_top">
              http://en.wikipedia.org/wiki/Wire-frame_model</a>
<br>or
<br> <a href="https://www.google.com/search?q=graphics+wireframe&tbm=isch" target="_top">
              https://www.google.com/search?q=graphics+wireframe&tbm=isch</a>
*/
public class Model
{
   public final List<Vertex> vertexList;
   public final List<LineSegment> lineSegmentList;
   public List<Color> colorList;

   public String  name;
   public boolean visible;
   public boolean debug;


   /**
      Construct an empty {@code Model} object.
   */
   public Model()
   {
      vertexList = new ArrayList<>();
      lineSegmentList = new ArrayList<>();
      colorList = new ArrayList<>();
      name = "";
      visible = true;
      debug = false;
   }


   /**
      Construct an empty {@code Model} object with the given name.

      @param name  a {link String} that is a name for this {@code Model}
   */
   public Model(String name)
   {
      this();
      this.name = name;
   }


   /**
      Construct a {@code Model} object with all the given data.

      @param vertexList       a {@link Vertex} {link List} for this {@code Model}
      @param lineSegmentList  a {@link LineSegment} {link List} for this {@code Model}
      @param colorList        a {@link Color} {link List} for this {@code Model}
      @param name             a {link String} that is a name for this {@code Model}
      @param visible          a {@code boolean} that determines this {@code Model}'s visibility
      @param debug            a {@code boolean} that turns debugging off/on for this {@code Model}
   */
   public Model(final List<Vertex> vertexList,
                final List<LineSegment> lineSegmentList,
                final List<Color> colorList,
                final String name,
                final boolean visible,
                final boolean debug)
   {
      this.vertexList = vertexList;
      this.lineSegmentList = lineSegmentList;
      this.colorList = colorList;
      this.name = name;
      this.visible = visible;
      this.debug = debug;
   }


   /**
      Add a {@link Vertex} (or vertices) to this {@code Model}'s
      {@link List} of vertices.

      @param vArray  array of {@link Vertex} objects to add to this {@code Model}
   */
   public final void addVertex(final Vertex... vArray)
   {
      for (final Vertex v : vArray)
      {
         vertexList.add( v );
      }
   }


   /**
      Get a {@link LineSegment} from this {@code Model}'s
      {@link List} of line segments.

      @param index  integer index of a {@link LineSegment} from this {@code Model}
      @return the {@link LineSegment} object at the given index
   */
   public final LineSegment getLineSegment(final int index)
   {
      return lineSegmentList.get(index);
   }


   /**
      Add a {@link LineSegment} (or LineSegments) to this {@code Model}'s
      {@link List} of line segments.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list. This method assumes that the appropriate vertices
      have been added to the {@code Model}'s {@link Vertex} list.

      @param lsArray  array of {@link LineSegment} objects to add to this {@code Model}
   */
   public final void addLineSegment(final LineSegment... lsArray)
   {
      for (final LineSegment ls : lsArray)
      {
         lineSegmentList.add(ls);
      }
   }


   /**
      Add a {@link Color} (or colors) to this {@code Model}'s
      {@link List} of colors.

      @param cArray  array of {@link Color} objects to add to this {@code Model}
   */
   public final void addColor(final Color... cArray)
   {
      for (final Color c : cArray)
      {
         this.colorList.add(c);
      }
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Model} object
   */
   @Override
   public String toString()
   {
      String result = "";
      result += "Model: " + name + "\n";
      result += "This Model's visibility is: " + visible + "\n";
      result += "Model has " + vertexList.size() + " vertices.\n";
      result += "Model has " + colorList.size() + " colors.\n";
      result += "Model has " + lineSegmentList.size() + " line segments.\n";
      int i = 0;
      for (Vertex v : this.vertexList)
      {
         result += i + ": " + v.toString();
         ++i;
      }
      //result = "Printing out this Model's " + colortList.size() + " colors:\n";
      i = 0;
      for (Color c : this.colorList)
      {
         result += i + ": " + c.toString() + "\n";
         ++i;
      }
      //result = "Printing out this Model's " + lineSegmentList.size() + " Line segments:\n";
      i = 0;
      for (LineSegment ls : this.lineSegmentList)
      {
         result += i + ": " + ls.toString();
         ++i;
      }
      //result += "Done printing out Model\n";
      return result;
   }
}
